with Vectors, Matrices, Params;
use Params;

pragma Elaborate_All (Vectors,Matrices,Params);

generic

  JMax: in Positive;          -- size of explicit chain
  Rho: in Radius;             -- for chain
  Rad: in Radius;             -- for CosSin1
  Psi1: in Potential := (others => Zero);
  Phi1: in Potential := (others => Zero);

  with procedure Proper_Rounding is <>;
  with package SM is new Matrices(Scalar => Scalar);

package CosSins1.Chain is

  package CSV is new Vectors (Component => CosSin1);

  JEMax:  constant Positive := JMax+1;
  subtype JRangeE is Integer range -JEMax .. JEMax;
  subtype CSVector is CSV.Vector;
  subtype CChain is CSVector(JRangeE); -- chain coefficients
  subtype SMatrix is SM.Matrix;

  type FChain is  --- bi-infinite chain of CosSin1 (plus a constant)
    record
      Sigma: Parity; --- 0: site-centered  1: bond-centered
      A: Scalar;     --- a constant, to be added to each C(J)
      C: CChain;
    end record;

  AIO: Boolean := False; --- read/write the component .A

  procedure SetZero(Par,Sigma: in Parity; Q: in out FChain);
  function CheckRad(Q: FChain) return Boolean;
  procedure FixRad(Q: in out FChain);
  function GetPar(Q: FChain; Check: Boolean := False) return Parity;
  procedure Copy(Q1: in FChain; Q2: in out FChain);
  function IsZero(Q: FChain) return Boolean;
  procedure Neg(Q: in out FChain);
  procedure Neg(Q1: in FChain; Q2: in out FChain);
  procedure Add(Q1: in FChain; Q2: in out FChain);
  procedure Sum(Q1,Q2: in FChain; Q3: in out FChain);
  procedure Sub(Q1: in FChain; Q2: in out FChain);
  procedure Diff(Q1,Q2: in FChain; Q3: in out FChain);
  procedure Mult(R: in Flt; Q: in out FChain);
  procedure Prod(R: in Flt; Q1: in FChain; Q2: in out FChain);
  function "*"(R: Flt; Q: FChain) return FChain;
  procedure Mult(S: in Scalar; Q: in out FChain);
  procedure Prod(S: in Scalar; Q1: in FChain; Q2: in out FChain);
  function "*"(S: Scalar; Q: FChain) return FChain;
  procedure AddProd(S: in Scalar; Q1: in FChain; Q2: in out FChain);
  procedure Prod(Q1,Q2: in FChain; Q3: in out FChain);
  procedure Norm(Q: in FChain; S: in out Scalar);
  function MaxNorm(Q: FChain) return Radius;
  procedure MaxNorms(Q: in FChain; R1,R2,R3: out Radius);
  procedure YRange(Q: in FChain; YInf,YSup: out Flt; Pow2: in Positive := 12);
  function Inf(Q: FChain; Pow2: Positive := 12) return Flt;
  function Sup(Q: FChain; Pow2: Positive := 12) return Flt;
  procedure NumTrunc(Siz: in Natural; Q1: in FChain; Q2: in out FChain);
  procedure AddBall(R: in Radius; Q: in out FChain);
  procedure AddBall(S: in Scalar; Q: in out FChain);
  procedure Component(J: in Integer; Q: in FChain; P: in out CosSin1);
  procedure Add_Site(J: in Integer; P: in CosSin1; Q: in out FChain);
  procedure Add_Tail(J: in Integer; P: in CosSin1; Q: in out FChain);
  procedure SetAverage0(Q: in out FChain);
  procedure Project(J: in Integer; Q: in out FChain);
  procedure Reflect(Q: in out FChain);
  procedure Reflect(Q1: in FChain; Q2: in out FChain);
  procedure TranslatePi(Q: in out FChain);
  procedure TranslatePi(Q1: in FChain; Q2: in out FChain);
  procedure Z_Symmetrize(T: in Symmetry; Q: in out FChain);
  procedure Pi_Symmetrize(T: in Symmetry; Q: in out FChain; Check: in Boolean := False);
  procedure Nabla(Q: in out FChain);
  procedure Nabla(Q1: in FChain; Q2: in out FChain);
  procedure NablaStar(Q: in out FChain);
  procedure NablaStar(Q1: in FChain; Q2: in out FChain);
  procedure Nabla1(Q: in out FChain);
  procedure Nabla1(Q1: in FChain; Q2: in out FChain);
  procedure NablaStar0(Q: in out FChain);
  procedure NablaStar0(Q1: in FChain; Q2: in out FChain);
  procedure Compose(C: in SPoly1; Q1: in FChain; Q2: in out FChain);
  procedure InvHelm(Omega,M: in Scalar; Q: in out FChain);
  procedure InvHelm(Omega,M: in Scalar; Q1: in FChain; Q2: in out FChain);
  procedure Show1(N: in String; P: in FChain; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; Q: in FChain; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; Q: in out FChain; Decimal: in Boolean := False);
  procedure Write(FileName: in String; Q: in FChain; Decimal: in Boolean := False);
  procedure Read(FileName: in String; Q: in out FChain; Decimal: in Boolean := False);

  procedure Phi2(Q: in FChain; Ph2: in out FChain; First: in Natural := 1); -- Phi"(Nabla1 Q)
  function Phi2(Q: FChain; First: Natural := 1) return FChain;
  procedure Psi2(Q: in FChain; Ps2: in out FChain; First: in Natural := 1); -- Psi"(Q)
  function Psi2(Q: FChain; First: Natural := 1) return FChain;
  procedure Phi2Range(Q: in FChain; YMin,YMax: out Flt; Pow2: in Positive := 12);
  procedure Psi2Range(Q: in FChain; YMin,YMax: out Flt; Pow2: in Positive := 12);

  --- modes etc

  type CMode is   -- a unit coeff or unit error
    record
      I:      EFreq;    -- CosSin1 index
      J:      JRangeE;  -- Chain index
      Par:    Parity;   -- 0: even in time   1: odd in time
      Sigma:  Parity;   -- 0: site-centered  1: bond-centered
      Err:    Boolean;  -- if True: J in JRange means CosSin1 error
      Simple: Boolean;  -- for maps that do not mess with error components
    end record;

  procedure Norm(M: in CMode; S: in out Scalar);
  procedure MaxNorm(M: in CMode; S: out Radius);
  procedure Assign(M: in CMode; C: in out CSVector);
  procedure Assign(M: in CMode; Q: in out FChain);
  procedure AddProd(M: in CMode; S: in Scalar; C: in out CSVector);
  procedure AddProd(M: in CMode; S: in Scalar; Q: in out FChain; Check: in Boolean);
  procedure Extract(M: in CMode; C: in out CSVector; S: in out Scalar);
  procedure Extract(M: in CMode; Q: in out FChain; S: in out Scalar; Check: in Boolean);
  procedure Check(M: in CMode);
  procedure Show(M: in CMode);

  type CModes is array (Positive range <>) of CMode; -- should start with index 1

  procedure Show(M: in CModes);
  procedure Check(M: in CModes);
  function Extract(J: Integer; M: CModes) return CModes;
  function Find(M: CModes; Par: Parity; I: Natural; J: Integer) return Natural;
  function Make(Siz: Positive; Par,Sigma: Parity; Simple,Trunc: Boolean) return CModes;

  --- numerical
  subtype PlotRange is Integer range -JMax .. JMax;

  procedure CenterRange(Q: in FChain; J1,J2: out PlotRange; Small: in Radius);
  procedure PutPlot(F: in File_Type; Q: in FChain; T: in Flt; J1,J2: in PlotRange; PutT,PutJ: in Boolean := False);
  procedure PutPlot(F: in File_Type; Q: in FChain; Steps: in Positive; J1,J2: in PlotRange; PutT,PutJ,Blank: in Boolean := False);
  procedure WritePlot(FileName: in String; Q: in FChain; Steps: in Positive);
  procedure WritePlot3d(FileName: in String; Q: in FChain; Steps: in Positive; Small: in Radius);

  -----------------------------------------------------
  -----------------------------------------------------

  RadPow:    constant RadVec := PowersOf(Rad,EDeg);       --- rounded up
  InvRad:    constant Radius := One/Rad;                  --- rounded up
  InvRadPow: constant RadVec := PowersOf(InvRad,EDeg);    --- rounded up
  Rho2:      constant Radius := Rho*Rho;                  --- rounded up
  RhoPow:    constant RadVec := PowersOf(Rho,2*JEMax);    --- rounded up
  InvRho:    constant Radius := One/Rho;                  --- rounded up
  InvRho2:   constant Radius := InvRho*InvRho;            --- rounded up
  InvRhoPow: constant RadVec := PowersOf(InvRho,2*JEMax); --- rounded up

  Prod_Parallel: Boolean := False;
  Comp_Parallel: Boolean := False;

  package FAlloc is new Recycling (Item => FChain);
  subtype LF_Pointer is FAlloc.LD_Pointer;
  FAC: FAlloc.Controlled_Protected_Data;
  FPool: FAlloc.Protected_Data renames FAC.P;

private

  procedure AddErrs(P1: in CosSin1; P2: in out CosSin1);
  procedure SumErrs(P1,P2: in CosSin1; P3: in out CosSin1);
  procedure AddProd(S: in Scalar; C1: in CChain; Q2: in out FChain);
  procedure Component(J: in Integer; Sigma: in Parity; C: in CChain; P: in out CosSin1);
  procedure NegNabla(Q: in out FChain);
  procedure NegNabla(Q1: in FChain; Q2: in out FChain);
  procedure NegNablaStar(Q: in out FChain);
  procedure NegNablaStar(Q1: in FChain; Q2: in out FChain);
  procedure Extract_Coeff(M: in CMode; C: in out CSVector; S: in out Scalar);
  procedure Extract_CosSin1_Error(M: in CMode; C: in out CSVector; S: in out Scalar);
  procedure Extract_FChain_Error(M: in CMode; C: in out CSVector; S: in out Scalar);
  function Compare(Sigma: Parity; J1,J2: Integer) return Integer;
  function SameType(M1,M2: CMode) return Boolean;
  procedure Check_Adjacent(M1,M2: in CMode);
  procedure ShowIJ(N: in String; M: in CMode; NewLine: in Boolean := True);
  procedure ShowType(M: in CMode);

  generic
    with function MaxDeg(N: Natural) return Integer;
  function Gen_Make(JCut: Positive; Par,Sigma: Parity; Simple,Include_Errs: Boolean) return CModes;

  JLast:    constant Positive := Choose(STrunc,JMax,JEMax);
  RhoPower: constant SVector  := PowersOf(Scal(Rho),2*JEMax);
  RadPower: constant SVector  := PowersOf(Scal(Rad),EDeg);

end CosSins1.Chain;
