with Strings, Globals, Protected_Counters, Flt_Vectors;
use Strings, Globals;

pragma Elaborate_All(Strings,Globals,Protected_Counters,Flt_Vectors);

package body MultiTaylors1.Phi is

  use SM,SV,STM,STV;

  procedure Show_Param is
  begin
    Show_Param_T1("S");
    Show1("M1.Phi Dim =",Dim);
  end Show_Param;

  function HSymmetry(Xh: TMatrix1) return Radius is
    Tmp: LT_Pointer;
    Err: Radius := Zero;
  begin
    TPool.Allocate(Tmp);
    declare
      P: Taylor1 renames Tmp.all.Data;
    begin
      for I in 1 .. Dim loop
        for J in 1 .. I-1 loop
          Diff(Xh(I+Dim,J),Xh(J+Dim,I),P);
          Err := RMax(MaxNorm(P),Err);
        end loop;
      end loop;
    end;
    TPool.Recycle(Tmp);
    return Err;
  end HSymmetry;

  function Symplecticity(U: Scalar_Matrix) return Radius is
    -- a norm of transposed(U)*J*U-J
    -- using formula (1.2) in the notes
    use Protected_Counters;
    L: constant Positive := U'Last(1)/2;
    NT: constant Natural := Reserve_Tasks(L-1,Flow_Parallel);
    PC: Protected_Counter(1);
    E: Flt_Vector(1 .. L);

    procedure Symp_Loop is
      I1,I2,J2,K2: Integer;
      AD,CB,AC,CA,DB,BD,E1,E2,E3,Tmp: Scalar;
    begin
      loop
        PC.Next(I1);
        exit when I1>L;
        I2 := I1+L;
        for J1 in 1 .. L loop
          J2 := J1+L;
          SetZero(AD); SetZero(CB);
          SetZero(AC); SetZero(CA);
          SetZero(DB); SetZero(BD);
          for K1 in 1 .. L loop
            K2 := K1+L;
            AddProd(U(K1,I1),U(K2,J2),AD,Tmp);
            AddProd(U(K2,I1),U(K1,J2),CB,Tmp);
            AddProd(U(K1,I1),U(K2,J1),AC,Tmp);
            AddProd(U(K2,I1),U(K1,J1),CA,Tmp);
            AddProd(U(K2,I2),U(K1,J2),DB,Tmp);
            AddProd(U(K1,I2),U(K2,J2),BD,Tmp);
          end loop;
          Diff(AD,CB,E1); -- transposed(A)*D-transposed(C)*B = I
          Diff(AC,CA,E2); -- transposed(A)*C-transposed(C)*A = 0
          Diff(DB,BD,E3); -- transposed(D)*B-transposed(B)*D = 0
          if J1=I1 then Sub(SOne,E1); end if;
          E(I1) := E(I1)+MaxNorm(E1)+MaxNorm(E2)+MaxNorm(E3);
        end loop;
      end loop;
    end Symp_Loop;

    task type Symp_Task_Type is end Symp_Task_Type;
    task body Symp_Task_Type is begin Symp_Loop; end Symp_Task_Type;

    use Flt_Vectors;
    Err: Radius := Zero;
  begin
    SetZero(E);
    if NT=0 then
      Symp_Loop;
    else
      declare
        Symp_Task: array(1 .. NT) of Symp_Task_Type;
      begin
        Symp_Loop;
      end;
      Free_Tasks(NT);
    end if;
    for I in 1 .. L loop Err := Err+E(I); end loop;
    return Err;
  end Symplecticity;

  -------------------- for LinFlow

  function MaxNorm(V: SVector) return Radius is
    --- upper bound on the norm
    S: Radius := Zero;
  begin
    for I in reverse V'Range loop
      S := S+MaxNorm(V(I));
    end loop;
    return S;
  end MaxNorm;

  procedure AddBall(R: in Radius; V: in out SVector) is
    S: Scalar;
  begin
    BallAt0(R,S);
    for I in V'Range loop Add(S,V(I)); end loop;
  end AddBall;

  procedure Quot(V: in SVector; R: in Flt; W: in out SVector) is
  begin
    for I in V'Range loop Quot(V(I),R,W(I)); end loop;
  end Quot;

  function OpNorm(A: SMatrix) return Radius is
    --- upper bound on operator norm
    S: Radius := Zero;
    SJ: Radius;
  begin
    for J in reverse A'Range(2) loop
      SJ := Zero;
      for I in reverse A'Range(1) loop
        SJ := SJ+MaxNorm(A(I,J));
      end loop;
      if SJ>S then S := SJ; end if;
    end loop;
    return S;
  end OpNorm;

  procedure LinFlow(X: in TMatrix; V: in out TVector; DT: in Radius; Tmp1,Tmp2: in out SVector) is
    use SSV_Ops;
    --- solves V'(t)=X(t)V(t) with given V(0)
    --- coefficient-errors can depend on t
    E,Contr: Flt;
    S: SVector renames Tmp1;
  begin
    Prod(X(0),V(0),V(1));
    for M in 1 .. PDeg-1 loop   -- compute V(M+1)
      Prod(X(0),V(M),S);
      for I in 1 .. M loop
        Prod(X(I),V(M-I),Tmp2);
        Add(Tmp2,S);
      end loop;
      Quot(S,Flt(M+1),V(M+1));
    end loop;
    if not STrunc then
      E := Zero;
      for M in reverse PDeg .. 2*PDeg loop
        Prod(X(M-PDeg),V(PDeg),S);
        for I in M-PDeg+1 .. PDeg loop
          Prod(X(I),V(M-I),Tmp2);
          Add(Tmp2,S);
        end loop;
        E := DT*E+MaxNorm(S)/Flt(M+1);
      end loop;                         -- E is degree PDeg+1 (or higher) error
      Contr := Zero;
      for M in reverse 0 .. PDeg loop
        Contr := DT*Contr+OpNorm(X(M));
      end loop;
      Contr := Contr*DT/Flt(PDeg+2);    -- contraction factor at degree D+1 (or higher)
      if Contr>=One then raise Sorry with "LinFlow: no contraction"; end if;
      E := E/abs(Contr-One);            -- assuming Flt is rounded up
      AddBall(DT*E,V(PDeg));
    end if;
  end LinFlow;

  procedure LinFlow(X: in TMatrix; V: in out TVecVec; DT: in Radius) is
    use Protected_Counters;
    NT: constant Natural := Reserve_Tasks(Dim2-1,Flow_Parallel);
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);

    procedure Flow_Loop is
      I: Natural;
      Tmp1,Tmp2: LM_Pointer;
    begin
      MPool.Allocate(Tmp1);
      MPool.Allocate(Tmp2);
      loop
        PC.Next(I);
        exit when I>Dim2;
        LinFlow(X,V(I),DT,Tmp1.all.Data,Tmp2.all.Data);
      end loop;
      MPool.Recycle(Tmp1);
      MPool.Recycle(Tmp2);
    end Flow_Loop;

    task type Flow_Task_Type is end Flow_Task_Type;
    task body Flow_Task_Type is begin
      Flow_Loop;
    exception
      when others => Task_Error := True; raise;
    end Flow_Task_Type;

  begin
    if NT=0 then
      Flow_Loop;
    else
      declare
        Flow_Task: array(1 .. NT) of Flow_Task_Type;
      begin
        Flow_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in MultiTaylors1.Phi.LinFlow"; end if;
    end if;
  end LinFlow;

  ------------------------

  procedure SetZero(V: in out TVector) is
  begin
    for M in Power loop SetZero(V(M)); end loop;
  end SetZero;

  procedure T2T(A: in out TMatrix1; B: in out TMatrix) is
    --- A is just input, contains garbage at the end
  begin
    for I in Index2X loop
      for J in Index2X loop
        declare
          AC: Poly1 renames A(I,J).C;
        begin
          for M in Power loop Swap(AC(M),B(M)(I,J)); end loop;
        end;
      end loop;
    end loop;
  end T2T;

  procedure MakeXh(M1,M2: in Scalar_Matrix; C: in Scalar_Vector; Time,DT: in Radius; Xh: in out TMatrix1) is
    --- components of the vector field X_H
    --- M1(I,*) are cosine coefficients of Alpha
    --- M2(I,*) are cosine coefficients of Beta
    STime: constant Scalar := Scal(Time);
    K2: Integer;
    U: Scalar;
    T: Taylor1;
  begin
    SetZero(Xh);
    for I in M1'Range(1) loop
      CosTaylor(I,STime,DT,T);
      for K in IndexX loop
        K2 := K+Dim;
        Copy(M2(I,K),U);
        if K>1 then
          Add(M1(I,K),U);
          AddProd(M1(I,K),T,Xh(K2,K-1));
        end if;
        if K<Dim then
          Add(M1(I,K+1),U);
          AddProd(M1(I,K+1),T,Xh(K2,K+1));
        end if;
        SubProd(U,T,Xh(K2,K));
      end loop;
    end loop;
    for K in IndexX loop
      Neg(C(K),U);
      Add(U,Xh(K+Dim,K));
    end loop;
  end MakeXh;

  function MakePhi0(DT: Radius) return TMatrix1 is
    pragma Unreferenced (DT);
    K2: Integer;
    P: TMatrix1;
  begin
    SetZero(P);
    for K1 in IndexX loop
      K2 := K1+Dim;
      AddCoeff(0,SOne,P(K1,K1));
      AddCoeff(0,SOne,P(K2,K2));
      AddCoeff(1,SOne,P(K1,K2));
    end loop;
    return P;
  end MakePhi0;

  function MakeInvPhi0(DT: Radius) return TMatrix1 is
    pragma Unreferenced (DT);
    K2: Integer;
    P: TMatrix1;
  begin
    SetZero(P);
    for K1 in IndexX loop
      K2 := K1+Dim;
      AddCoeff(0,SOne,P(K1,K1));
      AddCoeff(0,SOne,P(K2,K2));
      AddCoeff(1,SNegOne,P(K1,K2));
    end loop;
    return P;
  end MakeInvPhi0;

  procedure ValPhi0(T: in Scalar; V: in out SMatrix) is
    K2: Integer;
  begin
    SetZero(V);
    for K1 in IndexX loop
      K2 := K1+Dim;
      Copy(SOne,V(K1,K1));
      Copy(SOne,V(K2,K2));
      Copy(T,V(K1,K2));
    end loop;
  end ValPhi0;

  procedure Val(P: in TVector; X: in Scalar; Y: in out SVector) is
    --- assuming X in the domain
    use SSV_Ops;
  begin
    Copy(P(PDeg),Y);
    for I in reverse 0 .. PDeg-1 loop
      Mult(X,Y);
      Add(P(I),Y);
    end loop;
  end Val;

  procedure Run(DT: in Radius; M1,M2: in Scalar_Matrix; C: in Scalar_Vector; Phi00: in out SMatrix) is
    TwoPI: constant Scalar := Two*Pi;
    ScalDT: constant Scalar := Scal(DT);
    Ph0: constant TMatrix1 := MakePhi0(DT);
    Ph0i: constant TMatrix1 := MakeInvPhi0(DT);
    Time: Radius := Zero;
    T0: Scalar := SZero;
    U: SVector;
    ValPh0: SMatrix;
    Xh,XX,Tmp: TMatrix1;
    X: TMatrix;
    PhH: TVecVec;

    procedure MultVal(T: in Scalar) is
      use SSV_Ops;
    begin
      if not (T=T0) then
        ValPhi0(T,ValPh0);
        Copy(T,T0);
      end if;
      for I in Index2X loop
        Val(PhH(I),T,U);
        Mult(ValPh0,U);
        SetZero(PhH(I));
        Copy(U,PhH(I)(0));
      end loop;
    end MultVal;

    procedure Update_Phi00 is
      use SSV;
    begin
      for I in Index2X loop
        SetColumn(I,PhH(I)(0),Phi00);
      end loop;
      if Verbosity>0 and then Flt'Floor(Time)=Time then
        Show1("Time =",Time,False);
        Show1("  Symplectic Error =",Symplecticity(Phi00));
      end if;
    end Update_Phi00;

  begin
    Use_FBalls := True;
    STMProd_Parallel := Flow_Parallel;
    STVProd_Parallel := Flow_Parallel;
    if Verbosity>0 then Show1("DT =",DT); end if;
    for I in Index2X loop
      SetZero(PhH(I));
      Copy(SOne,PhH(I)(0)(I));
    end loop;
    loop
      MakeXh(M1,M2,C,Time,DT,Xh);
      if Verbosity>2 and then Flt'Floor(Time)=Time then
        Show1("Time =",Time,False);
        Show1("  H Symmetry Error =",HSymmetry(Xh));
      end if;
      Prod(Xh,Ph0,Tmp);
      Prod(Ph0i,Tmp,XX);
      T2T(XX,X);
      LinFlow(X,PhH,DT);
      if Scal(Time+DT)<TwoPI then
        MultVal(ScalDT);                          -- updates PhH
        Time := Time+DT;                          -- chosen to be representble
        if Verbosity>1 then Update_Phi00; end if; -- just for info
      else
        MultVal(TwoPI-Scal(Time));                -- updates PhH
        Time := Approx(TwoPI);                    -- just for info
        exit;
      end if;
    end loop;
    Update_Phi00;
  end Run;

end MultiTaylors1.Phi;
