with Globals, Strings, Ints, Newton, Quick_Sort;
use Globals, Strings, Ints;

pragma Elaborate_All (Globals,Strings,Ints,Newton,Quick_Sort);

package body PolyRoots is

  package Newton1 is new Newton(Numeric => STrunc, Scalar => Scalar, Operator => Scalar);
  package Newton2 is new Newton(Numeric => STrunc, Scalar => Vec2, Operator => Mat2);

  procedure Norm(X,Y: in Scalar; S,Tmp: in out Scalar) is
    --- S := Sqrt(X*X+Y*Y)
  begin
    if IsZero(Y) then
      Norm(X,S);
    elsif IsZero(X) then
      Norm(Y,S);
    else
      Prod(X,X,S);
      AddProd(Y,Y,S,Tmp);
      Norm(S,Tmp);
      Sqrt(Tmp,S);
    end if;
  end Norm;

  procedure Sqrt(Zr,Zi: in Scalar; Wr,Wi: in out Scalar) is
    --- W := Sqrt(Z) where Z=Zr+i*Zi and W=Wr+i*Wi
    --- branch chosen so that Wr>=0 and sign(Wi)=sign(Zi)
    Re_Sign: constant Integer := Sign(Zr);
    Im_Sign: constant Integer := Sign(Zi);
  begin
    if Im_Sign = 0 then      --- Z is real
      if Re_Sign >= 0 then
        Sqrt(Zr,Wr);
        SetZero(Wi);
      else
        Neg(Zr,Wr);
        Sqrt(Wr,Wi);
        SetZero(Wr);
      end if;
    else
      declare
        S,Tmp: Scalar;
      begin
        Norm(Zr,Zi,Tmp,S);
        Norm(Zr,S);
        Add(S,Tmp);
        Mult(Half,Tmp,S);
        Sqrt(Tmp,S);         --- S = Sqrt(Half*(Norm(Z)+abs(Zr)));
        if Re_Sign >= 0 then
          Copy(S,Wr);
        end if;
        if Im_Sign < 0 then
          Neg(S);
        end if;
        if Re_Sign <= 0 then
          Copy(S,Wi);
        end if;
        if Re_Sign < 0 then
          Quot(Zi,Wi,Wr);
          Mult(Half,Wr,Tmp);
        end if;
        if Re_Sign > 0 then
          Quot(Zi,Wr,Wi);
          Mult(Half,Wi,Tmp);
        end if;
      end;
    end if;
  end Sqrt;

  procedure QSolve(Zr,Zi: in Scalar; Wr1,Wi1,Wr2,Wi2: in out Scalar) is
    --- W=W1 and W=W2 solve W*W-2*Z*W+1=0
    Sr: Scalar renames Wr1;
    Si: Scalar renames Wi1;
    ZZr: Scalar renames Wr2;
    ZZi: Scalar renames Wi2;
    Tmp: Scalar renames Wi2;
  begin
    Prod(Zr,Zr,ZZr);
    Prod(Zi,Zi,Tmp);
    Sub(Tmp,ZZr);
    Prod(Zr,Zi,ZZi);
    Mult(Scal_Two,ZZi);  -- ZZ = Z*Z
    Sub(Scal_One,ZZr);
    Sqrt(ZZr,ZZi,Sr,Si); -- S = Sqrt(Z*Z-1)
    Diff(Zr,Sr,Wr2);
    Diff(Zi,Si,Wi2);     -- W2 = Z-Sqrt(Z*Z-1)
    Add(Zr,Wr1);
    Add(Zi,Wi1);         -- W1 = Z+Sqrt(Z*Z-1)
    if MaxNorm(Wr2)>MaxNorm(Wr1) then
      Swap(Wr1,Wr2);
      Swap(Wi1,Wi2);
    end if;
  end QSolve;

  --------------------------- basic
  --------------------------- basic

  function EffDeg(P: Poly) return Integer is
  begin
    for D in reverse 1 .. P'Last loop
      if not IsZero(P(D)) then return D; end if;
    end loop;
    return 0;
  end EffDeg;

  procedure Copy(P: in Poly; Q: in out Poly) is
  begin
    if Q'Last<P'Last then
      for I in 0 .. EffDeg(P) loop Copy(P(I),Q(I)); end loop;
    else
      for I in P'Range loop Copy(P(I),Q(I)); end loop;
      if Q'Last>P'Last then
        for I in P'Last+1 .. Q'Last loop SetZero(Q(I)); end loop;
      end if;
    end if;
  end Copy;

  procedure PolyVal(P: in Poly; Deg: in Natural; S: in Scalar; T: in out Scalar) is
  begin
    Copy(P(Deg),T);
    for I in reverse 0 .. Deg-1 loop
      Mult(S,T);
      Add(P(I),T);
    end loop;
  end PolyVal;

  procedure PolyVal(P: in Poly; S: in Scalar; T: in out Scalar) is
    Deg: constant Natural := EffDeg(P);
  begin
    Copy(P(Deg),T);
    for I in reverse 0 .. Deg-1 loop
      Mult(S,T);
      Add(P(I),T);
    end loop;
  end PolyVal;

  procedure PolyVal(P: in Poly; Sr,Si: in Scalar; Tr,Ti: in out Scalar) is
    --- P is assumed to be real
    Deg: constant Natural := EffDeg(P);
  begin
    SetZero(Ti);
    if IsZero(Si) then
      PolyVal(P,Deg,Sr,Tr);
    else
      Copy(P(Deg),Tr);
      declare
        Tmpr,Tmpi: Scalar;
      begin
        for I in reverse 0 .. Deg-1 loop
          Copy(Tr,Tmpr);
          Prod(Si,Ti,Tr);
          Neg(Tr);
          AddProd(Sr,Tmpr,Tr); -- Tr := Sr*Tr-Si*Ti
          Copy(Ti,Tmpi);
          Prod(Si,Tmpr,Ti);
          AddProd(Sr,Tmpi,Ti); -- Ti := Sr*Ti+Si*Tr
          Add(P(I),Tr);
        end loop;
      end;
    end if;
  end PolyVal;

  procedure PolyValDer(P: in Poly; Deg: in Natural; S: in Scalar; T: in out Scalar) is
  begin
    Prod(Scal(Flt(Deg)),P(Deg),T);
    for I in reverse 1 .. Deg-1 loop
      Mult(S,T);
      AddProd(Scal(Flt(I)),P(I),T);
    end loop;
  end PolyValDer;

  procedure PolyValDer(P: in Poly; S: in Scalar; T: in out Scalar) is
    Deg: constant Natural := EffDeg(P);
  begin
    Prod(Scal(Flt(Deg)),P(Deg),T);
    for I in reverse 1 .. Deg-1 loop
      Mult(S,T);
      AddProd(Scal(Flt(I)),P(I),T);
    end loop;
  end PolyValDer;

  procedure LMultiply(W: in Scalar; P: in out Poly) is
    --- multiply with (x-W)
    Deg: constant Natural := EffDeg(P);
    B: Scalar;
  begin
    Neg(W,B);
    for I in reverse 1 .. Deg+1 loop -- P(I) := P(I-1)-W*P(I)
      Mult(B,P(I));
      Add(P(I-1),P(I));
    end loop;
    Mult(B,P(0));
  end LMultiply;

  procedure LMultiply(A,B: in Scalar; P: in out Poly) is
    --- multiply with (A*x+B)
    Deg: constant Natural := EffDeg(P);
  begin
    if IsZero(A) then
      for I in 0 .. Deg loop Mult(B,P(I)); end loop;
    else
      for I in reverse 1 .. Deg+1 loop -- P(I) := A*P(I-1)+B*P(I)
        Mult(B,P(I));
        AddProd(A,P(I-1),P(I));
      end loop;
      Mult(B,P(0));
    end if;
  end LMultiply;

  procedure QMultiply(U,V,C,D: in Scalar; P: in out Poly) is
    --- multiply with (x*x+U*x+V) and then add (C*x+D)
    Deg: constant Natural := EffDeg(P);
  begin
    for I in reverse 2 .. Deg+2 loop -- P(I) := P(I-2)+U*P(I-1)+V*P(I);
      Mult(V,P(I));
      AddProd(U,P(I-1),P(I));
      Add(P(I-2),P(I));
    end loop;
    Mult(V,P(1));
    AddProd(U,P(0),P(1));
    Mult(V,P(0));
    Add(D,P(0));
    Add(C,P(1));
  end QMultiply;

  procedure QProduct(U,V,C,D: in Scalar; Q: in Poly; P: in out Poly) is
  begin
    Copy(Q,P);
    QMultiply(U,V,C,D,P);
  end QProduct;

  procedure QMultiply(U,V: in Vector; A,B: in Scalar; P: in out Poly) is
  begin
    for I in P'Range loop SetZero(P(I)); end loop;
    Assign(One,P(0));
    for K in U'Range loop
      QMultiply(U(K),V(K),Scal_Zero,Scal_Zero,P);
    end loop;
    LMultiply(A,B,P);
  end QMultiply;

  procedure Roots2Poly(W: in Vector; P: in out Poly) is
  begin
    for I in P'Range loop SetZero(P(I)); end loop;
    Assign(One,P(0));
    for I in W'Range loop LMultiply(W(I),P); end loop;
  end Roots2Poly;

  procedure Roots2Poly(Wr,Wi: in Vector; P: in out Poly) is
    --- assuming cc roots are next to each other
    Scal_NegTwo: constant Scalar := Scal(-Two);
    I: Integer := 1;
    U,V,S: Scalar;
  begin
    for J in P'Range loop SetZero(P(J)); end loop;
    Assign(One,P(0));
    while I <= Wr'Last loop
      if IsZero(Wi(I)) then
        LMultiply(Wr(I),P);
        I := I+1;
      else
        Diff(Wr(I),Wr(I+1),S);
        if not IsZero(S) then
          Show2("Wr ",Wr(I),Wr(I+1));
          raise Not_Implemented;
        end if;
        Sum(Wi(I),Wi(I+1),S);
        if not IsZero(S) then
          Show2("Wr ",Wi(I),Wi(I+1));
          raise Not_Implemented;
        end if;
        Prod(Scal_NegTwo,Wr(I),U);
        Prod(Wr(I),Wr(I),V);
        AddProd(Wi(I),Wi(I),V);
        QMultiply(U,V,Scal_Zero,Scal_Zero,P);
        I := I+2;
      end if;
    end loop;
  end Roots2Poly;

  procedure Root(P0,P1: in Scalar; W: in out Scalar) is
  begin
    if IsZero(P1) then
      Show0("PolyRoots.Root: degree 0");
      raise Not_Found;
    end if;
    Quot(P0,P1,W);
    Neg(W);
  end Root;

  procedure RealRoots2(P0,P1,P2: in Scalar; W1,W2: in out Scalar) is
    B,C,V: Scalar;
  begin
    Quot(P1,P2,B);
    Quot(P0,P2,C);
    Roots2(B,C,W1,W2,V);
    if not IsZero(V) then
      Show0("PolyRoots.RealRoots2: no real roots found");
      raise Not_Found;
    end if;
  end RealRoots2;

  procedure FindRealRoot(P: in Poly; W: in out Scalar; Err: out Radius) is
    Deg: constant Natural := EffDeg(P);
  begin
    if Deg<3 then
      if Deg=2 then
        declare
          W1,W2,S1,S2: Scalar;
        begin
          RealRoots2(P(0),P(1),P(2),W1,W2);
          Diff(W1,W,S1);
          Diff(W2,W,S2);
          if MaxNorm(S1)<MaxNorm(S2) then
            Copy(W1,W);
          else
            Copy(W2,W);
          end if;
          Err := Zero;
        end;
      else
        Root(P(0),P(1),W);
        Err := Zero;
      end if;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;

        procedure TF(X: in Scalar; FX,DFX: in out Scalar) is
        begin
          PolyVal(P,Deg,X,FX);
          PolyValDer(P,Deg,X,DFX);
          Err := MaxNorm(FX);
        end TF;

        use Newton1;
        procedure Find is new Find0 (TF => TF);
      begin
        Find(W,Steps,Verb);
      exception
        when No_Contraction =>
          Show0("PolyRoots.FindRealRoot: no contraction");
          raise Not_Found;
      end;
      if Verbosity>4 then Show1("FindRealRoot Err =",Err); end if;
    end if;
  end FindRealRoot;

  procedure FindRealRoots(P: in Poly; W: in out Vector) is
    --- dies if there is a problem
    D: Integer := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
  begin
    if W'First /= 1 then raise Not_Implemented; end if;
    if D<3 then
      if D=2 then
        RealRoots2(P(0),P(1),P(2),W(1),W(2));
      else
        Root(P(0),P(1),W(1));
      end if;
    else
      for N in W'Range loop
        if D>0 then
          FindRealRoot(P,W(N),Err);
          if Err>MaxErr then MaxErr := Err; end if;
        end if;
        D := D-1;
      end loop;
      if D<0 and then Verbosity>1 then
        Show2("FindRealRoots warning: W'Range too large:",W'First,W'Last);
      end if;
    end if;
    if Verbosity>3 then
      Show1("FindRealRoots: MaxErr =",MaxErr);
    end if;
  end FindRealRoots;

  procedure LDivision(P: in out Poly; W: in Scalar; R: in out Scalar) is
    Deg: constant Natural := EffDeg(P);
    Pt: Scalar renames P(Deg);
  begin
    if Deg=0 then
      Copy(W,R);
    else
      Swap(Pt,P(Deg-1));
      for I in reverse 1 .. Deg-1 loop  -- Q(I-1) := W*Q(I)+P(I)
        Swap(Pt,P(I-1));
        AddProd(W,P(I),P(I-1));
      end loop;
      Prod(W,P(0),R);
      Add(Pt,R);
      SetZero(Pt);
    end if;
  end LDivision;

  procedure LDivision(P: in Poly; W: in Scalar; Q: in out Poly; R: in out Scalar) is
  begin
    Copy(P,Q);
    LDivision(Q,W,R);
  end LDivision;

  procedure QDivision(P: in Poly; U,V: in Scalar; Q: in out Poly; C,D: in out Scalar) is
    S: constant Vec2 := (U,V);
    FS: Vec2;
  begin
    Bairstow0(P,EffDeg(P),S,Q,FS);
    Copy(FS(1),C);
    Copy(FS(2),D);
  end QDivision;

  procedure QDivision(P: in Poly; U,V: in Scalar; Q: in out Poly) is
    S: constant Vec2 := (U,V);
    FS: Vec2;
  begin
    Bairstow0(P,EffDeg(P),S,Q,FS);
  end QDivision;

  procedure Bairstow0(P: in Poly; Deg: in Integer; S: in Vec2; Q: in out Poly; FS: in out Vec2) is
    --- Bairstow's method (wikipedia notation)
    --- P(x)=(x*x+U*x+V)*Q(x)+(C*x+D)
    U: Scalar renames S(1);
    V: Scalar renames S(2);  -- x*x+U*x+V
    C: Scalar renames FS(1);
    D: Scalar renames FS(2); -- C*x+D
  begin
    if Deg<2 then
      for I in Q'Range loop SetZero(Q(I)); end loop;
      Copy(P(1),C);
      Copy(P(0),D);
    else
      for I in reverse Deg-1 .. Q'Last loop
        SetZero(Q(I));
      end loop;
      Copy(P(Deg),Q(Deg-2));
      if Deg>2 then
        Neg(P(Deg-1),Q(Deg-3));
        AddProd(U,Q(Deg-2),Q(Deg-3));
        Neg(Q(Deg-3));
        for I in reverse 0 .. Deg-4 loop
          Neg(P(I+2),Q(I));
          AddProd(V,Q(I+2),Q(I));
          AddProd(U,Q(I+1),Q(I));
          Neg(Q(I));
        end loop;
      end if;
      Neg(P(1),C);
      AddProd(U,Q(0),C);
      AddProd(V,Q(1),C);
      Neg(C);
      Neg(P(0),D);
      AddProd(V,Q(0),D);
      Neg(D);
    end if;
  end Bairstow0;

  procedure Bairstow1(Q: in Poly; Deg: in Integer; S: in Vec2; R: in out Poly; DFS: in out Mat2) is
    --- derivative for Bairstow0
    T: Vec2;
    U: Scalar renames S(1);
    V: Scalar renames S(2);  -- x*x+U*x+V
    G: Scalar renames T(1);
    H: Scalar renames T(2);
  begin
    Bairstow0(Q,Deg,S,R,T);
    Prod(G,U,DFS(1,1));
    Sub(H,DFS(1,1));    --  GU-H
    Neg(G,DFS(1,2));    --         -G
    Prod(G,V,DFS(2,1)); --  GV
    Neg(H,DFS(2,2));    --         -H
  end Bairstow1;

  procedure FindQFactor2(P0,P1,P2: in Scalar; U,V: in out Scalar; Q: in out Poly) is
  begin
    if IsZero(P2) then
      Show0("PolyRoots.FindQFactor2: no quadratic factor");
      raise Not_Found;
    end if;
    for I in 1 .. Q'Last loop SetZero(Q(I)); end loop;
    Copy(P2,Q(0));
    Quot(P1,P2,U);
    Quot(P0,P2,V);
  end FindQFactor2;

  procedure FindQFactor(P: in Poly; U,V: in out Scalar; Q: in out Poly; Err: out Radius) is
    --- U,V are approximate solutions
    --- if not STrunc then only error estimates are added
    DegP: constant Natural := EffDeg(P);
  begin
    if DegP<3 then
      FindQFactor2(P(0),P(1),P(2),U,V,Q);
      Err := Zero;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;
        DegQ: constant Natural := DegP-2;
        DegR: constant Integer := DegP-4;
        R: Poly(0 .. IMax(DegR,1));

        procedure TF(S: in Vec2; FS: in out Vec2; DFS: in out Mat2) is
        begin
          Bairstow0(P,DegP,S,Q,FS);
          Bairstow1(Q,DegQ,S,R,DFS);
          Err := MaxNorm(FS);
        end TF;

        S: Vec2 := (U,V);

        use Newton2;
        procedure Find is new Find0 (TF => TF);
      begin
        Find(S,Steps,Verb);   --- last step produces correct Q
        Copy(S(1),U);
        Copy(S(2),V);
      exception
        when No_Contraction =>
          Show0("PolyRoots.FindQFactor: no contraction");
          raise Not_Found;
      end;
    end if;
  end FindQFactor;

  procedure FindQFactors(P: in Poly; U,V: in out Vector; A,B: in out Scalar) is
    --- last factor is A*x+B
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
    S: Scalar;
    Q: Poly(0 .. Deg);
  begin
    for N in 1 .. Deg/2 loop
      FindQFactor(P,U(N),V(N),Q,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      if Verbosity>4 then
        Show1(Strng(N,4) & " FindQFactors Err =",Err);
      end if;
    end loop;
    if (Deg mod 2)=1 then
      Quot(B,A,S);
      Neg(S);
      FindRealRoot(P,S,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      Assign(One,A);
      Neg(S,B);
    else
      SetZero(A);
      Assign(One,B);
    end if;
    if Verbosity>3 then
      Show1("FindRealRoots: MaxErr =",MaxErr);
    end if;
  end FindQFactors;

  procedure FindRoots(P: in Poly; Wr,Wi: in out Vector; D: in Natural) is
    --- assuming cc roots are consecutive in Wr,Wi
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    I: Integer := 1;
    Err: Radius;
    U,V,S: Scalar;
    Q: Poly(0 .. Deg);
  begin
    while I <= D loop
      if IsZero(Wi(I)) then
        FindRealRoot(P,Wr(I),Err);      --- Newton1
        I := I+1;
      else
        Sum(Wr(I),Wr(I+1),U);
        Diff(Wi(I),Wi(I+1),S);
        Prod(U,U,V);
        AddProd(S,S,V);
        Mult(Quarter,V);
        Neg(U);
        FindQFactor(P,U,V,Q,Err);       --- Newton2
        Roots2(U,V,Wr(I),Wr(I+1),Wi(I));
        Neg(Wi(I),Wi(I+1));
        I := I+2;
      end if;
    end loop;
  end FindRoots;

  ------------------------------------ symplectic
  ------------------------------------ symplectic

  procedure Clenshaw(C: in Cheby; P: in out Poly) is
    --- convert a finite Cheby series to a polynomial
    Deg: constant Natural := EffDeg(C);
  begin
    if Deg<2 then
      Copy(C,P);
    else
      declare
        B2: Poly renames P;
        B1: Poly(0 .. Deg);
      begin
        for I in B1'Range loop SetZero(B1(I)); end loop;
        for I in B2'Range loop SetZero(B2(I)); end loop;
        Copy(C(Deg),B1(0));
        for K in reverse 1 .. Deg-1 loop
          for I in 1 .. Deg-K loop
            Neg(B2(I));
            AddProd(Scal_Two,B1(I-1),B2(I));
          end loop;
          Neg(B2(0));
          Add(C(K),B2(0));
          for I in 0 .. Deg-K loop Swap(B1(I),B2(I)); end loop;
        end loop;
        for I in 1 .. Deg loop
          Neg(B2(I));
          Add(B1(I-1),B2(I));
        end loop;
        Neg(B2(0));
        Add(C(0),B2(0));
      end;
    end if;
  end Clenshaw;

  procedure SymPolyCheby(P: in Poly; C: in out Cheby) is
    Deg: constant Natural := EffDeg(P);
    M: constant Natural := Deg/2;
  begin
    if 2*M /= Deg then raise Undefined; end if;
    Copy(P(M),C(0));
    for K in 1 .. M loop
      Prod(Scal_Two,P(M-K),C(K));
    end loop;
    for K in M+1 .. C'Last loop
      SetZero(C(K));
    end loop;
  end SymPolyCheby;

  procedure SymCosPoly(P: in Poly; Q: in out Poly) is
    --- a factor (x*x+U*x+1) of P corresponds to a root -U/2 for Q
    M: constant Natural := EffDeg(P)/2;
    C: Cheby(0 .. M);
  begin
    SymPolyCheby(P,C);
    Clenshaw(C,Q);
  end SymCosPoly;

  function Right_Sign(P: Poly) return Integer is
    S0: constant Integer := Sign(P(0));
    T: Scalar := P(0);
  begin
    for I in 1 .. P'Last loop Add(P(I),T); end loop; -- value at 1
    return Sign(T)*S0;
  end Right_Sign;

  function Left_Sign(P: Poly) return Integer is
    S0: constant Integer := Sign(P(0));
    Odd: Boolean := True;
    T: Scalar := P(0);
  begin
    for I in 1 .. P'Last loop
      if Odd then Sub(P(I),T); else Add(P(I),T); end if;
      Odd := not Odd;
    end loop;           -- value at -1
    return Sign(T)*S0;
  end Left_Sign;

  procedure CircleRoots(P: in Poly; Hr: in out Vector) is
    --- roots of P need to be on the unit circle
    --- dies if there is a problem
    Deg: constant Positive := EffDeg(P);
    Q: Poly(0 .. Deg/2);
  begin
    if (Deg mod 2)=1 then raise Not_Implemented; end if;
    if Left_Sign(P)<0 then raise Off_Circle; end if;
    if Right_Sign(P)<0 then raise Off_Circle; end if;
    SymCosPoly(P,Q);
    FindRealRoots(Q,Hr);
  end CircleRoots;

  ------------------------ older

  procedure FindSymQFactor2(P0,P1,P2: in Scalar; U: in out Scalar; Q: in out Poly) is
    --- for degree 2
  begin
    if IsZero(P2) then
      Show0("PolyRoots.FindSymQFactor2: no quadratic factor");
      raise Not_Found;
    end if;
    if Verbosity>2 then
      Quot(P0,P2,U);
      Sub(Scal_One,U);
      Show1("FindSymQFactor2: last |V-1| =",MaxNorm(U));
    end if;
    for I in 1 .. Q'Last loop SetZero(Q(I)); end loop;
    Copy(P2,Q(0));
    Quot(P1,P2,U);
  end FindSymQFactor2;

  procedure FindSymQFactor(P: in Poly; U0: in out Scalar; Q: in out Poly; Err: out Radius) is
    --- roots are assumed to come in pairs (r,1/r)
    --- U0 is an approximate solution (V=1)
    --- if not STrunc then only error estimates are added
    DegP: constant Natural := EffDeg(P);
  begin
    if DegP<3 then
      FindSymQFactor2(P(0),P(1),P(2),U0,Q);
      Err := Zero;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;
        DegQ: constant Natural := DegP-2;
        DegR: constant Integer := DegP-4;
        S,FS: Vec2;
        DFS: Mat2;
        U: Scalar renames S(1);
        V: Scalar renames S(2);
        C: Scalar renames FS(1);
        D: Scalar renames FS(2);
        DC: Scalar renames DFS(1,1);
        R: Poly(0 .. IMax(DegR,1));

        procedure TF(X: in Scalar; FX,DFX: in out Scalar) is
        begin
          Copy(X,U);
          Bairstow0(P,DegP,S,Q,FS);
          Copy(C,FX);
          Bairstow1(Q,DegQ,S,R,DFS);
          Copy(DC,DFX);
        end TF;

        use Newton1;
        procedure Find is new Find0 (TF => TF);
      begin
        Assign(One,V);
        Find(U0,Steps,Verb);  --- last step yields correct Q
        Err := MaxNorm(FS);
        if Verbosity>2 then
          Show2("FindSymQFactor: C D ",C,D);
        end if;
      exception
        when No_Contraction =>
          Show0("PolyRoots.FindSymQFactor: no contraction");
          raise Not_Found;
      end;
    end if;
  end FindSymQFactor;

  procedure FindSymQFactors(P: in Poly; U: in out Vector) is
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
    Q: Poly(0 .. Deg);
  begin
    if (Deg mod 2)=1 then raise Undefined; end if;  --- Deg needs to be even
    for N in 1 .. Deg/2 loop
      FindSymQFactor(P,U(N),Q,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      if Verbosity>4 then
        Show1(Strng(N,4) & " Symm_FindQFactors Err =",Err);
      end if;
    end loop;
    if Verbosity>3 then
      Show1("FindSymQFactors: MaxErr =",MaxErr);
    end if;
  end FindSymQFactors;

  ------------------------------------ sorting
  ------------------------------------ sorting

  function GT_MaxNorm(A,B: Scalar) return Boolean is
  begin
    return MaxNorm(A)>MaxNorm(B);
  end GT_MaxNorm;

  procedure Dec_MaxNorm_Sort(W: in out Vector) is
    procedure QSort is new Quick_Sort (Item => Scalar, Item_Vector => Vector, "<" => GT_MaxNorm);
  begin
    QSort(W'First,W'Last,W);
  end Dec_MaxNorm_Sort;

  function GT_MaxNorm(Eps: Radius; Wr,Wi: Vector; I,J: Integer) return Boolean is
    OneMinusEps: constant Radius := One-Eps;
    NrI: constant Flt := MaxNorm(Wr(I));
    AiI: constant Flt := Approx(Wi(I));
    NI: constant Flt := NrI*NrI+AiI*AiI;
    NrJ: constant Flt := MaxNorm(Wr(J));
    AiJ: constant Flt := Approx(Wi(J));
    NJ: constant Flt := NrJ*NrJ+AiJ*AiJ;
  begin
    if OneMinusEps*NI > NJ then return  True; end if;
    if OneMinusEps*NJ > NI then return  False; end if;
    if OneMinusEps*NrI > NrJ then return  True; end if;
    if OneMinusEps*NrJ > NrI then return  False; end if;
    return AiI>AiJ;
  end GT_MaxNorm;

  procedure Dec_MaxNorm_Sort(Eps: in Radius; Wr,Wi: in out Vector) is
    K: Integer;
    procedure Swap(I,J: in out Integer) is begin K := I; I := J; J := K; end Swap;
    function GT(I,J: Integer) return Boolean is begin return GT_MaxNorm(Eps,Wr,Wi,I,J); end GT;
    procedure QSort is new Quick_Sort (Item => Integer, Item_Vector => Int_Vec, "<" => GT);
    P: Int_Vec(Wr'Range); -- Pivot
    Tmp: Vector(Wr'Range);
  begin
    for I in P'Range loop P(I) := I; end loop;
    QSort(P'First,P'Last,P);
    for I in Wr'Range loop Copy(Wr(I),Tmp(I)); end loop;
    for I in Wr'Range loop Copy(Tmp(P(I)),Wr(I)); end loop;
    for I in Wi'Range loop Copy(Wi(I),Tmp(I)); end loop;
    for I in Wi'Range loop Copy(Tmp(P(I)),Wi(I)); end loop;
  end Dec_MaxNorm_Sort;

  ------------------------------------ numerical
  ------------------------------------ numerical

  procedure RoughRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Err: in out Radius) is
    Deg: constant Natural := EffDeg(P);
    W1,W2: Scalar;
  begin
    if Deg<3 then
      if Deg=2 then
        RealRoots2(P(0),P(1),P(2),W1,W2);
        if MaxNorm(W1)>MaxNorm(W2) then
          Copy(W1,W);
        else
          Copy(W2,W);
        end if;
      else
        Root(P(0),P(1),W);
      end if;
      Err := Zero;
    else
      declare
        X1: Flt := WMin;
        X2: Flt := WMax;
        Y1,Y2: Flt;
        FW: Scalar;

        function RF(R: Flt) return Flt is
        begin
          Assign(R,W);
          PolyVal(P,Deg,W,FW);
          return Approx(FW);
        end RF;

        function SF(Wt: Scalar) return Flt is
        begin
          PolyVal(P,Deg,Wt,FW);
          return Approx(FW);
        end SF;

        procedure BFindSigns is new Bisect.Find_Signs (F => RF);
        procedure BFindZero is new Bisect.FindZero (F => SF);
      begin
        BFindSigns(X1,X2,Y1,Y2);
        Assign(X1,W1);
        Assign(X2,W2);
        if Y1*Y2<Zero then
          if Y2<Zero then Swap(W1,W2); end if;
          BFindZero(W1,W2,Err);
          Sum(W1,W2,W);
          Mult(Scal_Half,W);
        elsif Y1=Zero then
          Copy(W1,W);
          Center(W1);
          Sub(W,W1);
          Err := MaxNorm(W1);
        elsif Y2=Zero then
          Copy(W2,W);
          Center(W2);
          Sub(W,W2);
          Err := MaxNorm(W2);
        else
          Show1("PolyRoots.RoughRealRoot: no sign change Found. Deg =",Deg);
          raise Not_Found;
        end if;
      end;
    end if;
    if Verbosity>4 then
      Show1("RoughRealRoot Err =",Err);
    end if;
  end RoughRealRoot;

  procedure NumRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius) is
    Err: Radius := Eps;
  begin
    RoughRealRoot(P,WMin,WMax,W,Err);
    FindRealRoot(P,W,Err);
  end NumRealRoot;

  procedure NumRealRoots(P: in Poly; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius) is
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    Q: Poly(0 .. Deg);
  begin
    Copy(P,Q);
    if Deg<2 then --- ignore WMin,WMax
      if Deg=0 then raise Undefined; end if;
      Quot(P(0),P(1),W(1));
      Neg(W(1));
      Swap(Q(1),Q(0));
      SetZero(Q(1));
    else
      declare
        Err: Radius;
        R: Scalar;
      begin
        for N in 1 .. Deg loop
          Err := Eps;
          RoughRealRoot(Q,WMin,WMax,W(N),Err);
          if Err>Eps then
            if Verbosity>1 then Show1("NumRealRoots: Rough Err =",Err); end if;
            SetZero(W(N));
            return;
          end if;
          FindRealRoot(Q,W(N),Err);
          if Err>Eps then
            if Verbosity>1 then
              Show1("NumRealRoots: Find Err =",Err);
            end if;
            SetZero(W(N));
            return;
          end if;
          LDivision(Q,W(N),R);
        end loop;
      end;
    end if;
    FindRealRoots(P,W);
  end NumRealRoots;

  procedure NumCircleRoots(P: in Poly; Hr: in out Vector; Eps: in Radius) is
    --- roots of P need to be on the unit circle
    Deg: constant Positive := EffDeg(P);
    Q: Poly(0 .. Deg/2);
  begin
    if (Deg mod 2)=1 then raise Not_Implemented; end if;
    SymCosPoly(P,Q);
    NumRealRoots(Q,NegOne,One,Hr,Eps);
  end NumCircleRoots;

  procedure RoughSymQFactor(P: in Poly; UMin,UMax: in Flt; U0: in out Scalar; Tmp: in out Poly; Err: in out Radius) is
    Deg: constant Natural := EffDeg(P);
    Q: Poly renames Tmp;
  begin
    if Deg<3 then
      FindSymQFactor2(P(0),P(1),P(2),U0,Q);
      Err := Zero;
    else
      declare
        X1: Flt := UMin;
        X2: Flt := UMax;
        Y1,Y2: Flt;
        U1,U2: Scalar;
        S,FS: Vec2;
        U: Scalar renames S(1);
        C: Scalar renames FS(1);

        function RF(Ut: Flt) return Flt is
        begin
          Assign(Ut,U);
          Bairstow0(P,Deg,S,Q,FS);
          return Approx(C);
        end RF;

        function SF(Ut: Scalar) return Flt is
        begin
          Copy(Ut,U);
          Bairstow0(P,Deg,S,Q,FS);
          return Approx(C);
        end SF;

        procedure BFindSigns is new Bisect.Find_Signs (F => RF);
        procedure BFindZero is new Bisect.FindZero (F => SF);
      begin
        Assign(One,S(2));
        BFindSigns(X1,X2,Y1,Y2);
        Assign(X1,U1);
        Assign(X2,U2);
        if Y1*Y2<Zero then
          if Y2<Zero then Swap(U1,U2); end if;
          BFindZero(U1,U2,Err);
          Sum(U1,U2,U);
          Mult(Scal_Half,U);
        elsif Y1=Zero then
          Copy(U1,U);
          Center(U1);
          Sub(U,U1);
          Err := MaxNorm(U1);
        elsif Y2=Zero then
          Copy(U2,U);
          Center(U2);
          Sub(U,U2);
          Err := MaxNorm(U2);
        else
          Show1("PolyRoots.RoughSymQFactor: no sign change Found. Deg =",Deg);
          raise Not_Found;
        end if;
        Copy(U,U0);
      end;
    end if;
    if Verbosity>4 then
      Show1("RoughSymQFactor Err =",Err);
    end if;
  end RoughSymQFactor;

  procedure NumSymQFactor(P: in Poly; UMin,UMax: in Flt; U: in out Scalar; Eps: in Radius) is
    Deg: constant Natural := EffDeg(P);
    Err: Radius := Eps;
    Q: Poly(0 .. Deg-2);
  begin
    RoughSymQFactor(P,UMin,UMax,U,Q,Err);
    FindSymQFactor(P,U,Q,Err);
  end NumSymQFactor;

  procedure NumSymQFactors(P: in Poly; UMin,UMax: in Flt; U: in out Vector; Eps: in Radius) is
    Deg: constant Natural := EffDeg(P);
    Err: Radius := Eps;
    Pt,Q: Poly(0 .. Deg);
  begin
    Copy(P,Q);
    if (Deg mod 2)=1 then raise Undefined; end if;
    for N in 1 .. Deg/2 loop
      Copy(Q,Pt);
      Err := Eps;
      RoughSymQFactor(Pt,UMin,UMax,U(N),Q,Err);
      if Err>Eps then
        if Verbosity>1 then Show1("NumSymQFactors: Rough Err =",Err); end if;
        SetZero(U(N));
        return;
      end if;
      FindSymQFactor(Pt,U(N),Q,Err);
      if Err>Eps then
        if Verbosity>1 then Show1("NumSymQFactors: Find Err =",Err); end if;
        SetZero(U(N));
        return;
      end if;
    end loop;
    FindSymQFactors(P,U);
  end NumSymQFactors;

  ------------------------------------- testing stuff
  ------------------------------------- testing stuff

  procedure Product(P,Q: in Poly; R: in out Poly) is
    DP: constant Natural := EffDeg(P);
    DQ: constant Natural := EffDeg(Q);
  begin
    for K in R'Range loop SetZero(R(K)); end loop;
    for I in 0 .. DP loop
      for J in 0 .. DQ loop
        AddProd(P(I),Q(J),R(I+J));
      end loop;
    end loop;
  end Product;

  procedure SymProd(P1: in Poly; P2: in out Poly) is
    D: constant Natural := EffDeg(P1);
  begin
    for K in P2'Range loop SetZero(P2(K)); end loop;
    for I in 0 .. D loop
      for J in 0 .. D loop
        AddProd(P1(D-I),P1(J),P2(I+J));
      end loop;
    end loop;
  end SymProd;

  procedure Chebyshev(N: in Natural; P: in out Poly) is
  begin
    if N<2 then
      for I in P'Range loop SetZero(P(I)); end loop;
      Assign(One,P(N));
    else
      declare
        C: Cheby(0 .. N);
      begin
        for I in 0 .. N-1 loop SetZero(C(I)); end loop;
        Assign(One,C(N));
        Clenshaw(C,P);
      end;
    end if;
  end Chebyshev;

  function SymErr(P: Poly) return Radius is
    Deg: constant Natural := EffDeg(P);
    M: constant Natural := Deg/2;
    R: Radius := Zero;
    S: Scalar;
  begin
    if 2*M /= Deg then raise Undefined; end if;
    for K in reverse 1 .. M loop
      Diff(P(M+K),P(M-K),S);
      R := RMax(MaxNorm(S),R);
    end loop;
    return R;
  end SymErr;

  procedure RandomRoots(W: in out Vector; Fac: in Flt := One) is
  begin
    for I in W'Range loop Assign(Fac*Simple_Random,W(I)); end loop;
  end RandomRoots;

  procedure RandomRoots(NSimple: in Natural; Wr,Wi: in out Vector; RFac,IFac: in Flt := One) is
    Deg: constant Positive := Wr'Last;
    I: Integer := 1;
  begin
    loop
      if I <= NSimple then
        Assign(RFac*Simple_Random,Wr(I));
        SetZero(Wi(I));
        I := I+1;
      else
        Assign(RFac*Simple_Random,Wr(I));
        Copy(Wr(I),Wr(I+1));
        Assign(IFac*Simple_Random,Wi(I));
        Neg(Wi(I),Wi(I+1));
        I := I+2;
      end if;
      exit when I>Deg;
    end loop;
  end RandomRoots;

  procedure RandomReal(W: in out Vector; P: in out Poly; Fac: in Flt := One) is
  begin
    RandomRoots(W,Fac);
    Roots2Poly(W,P);
  end RandomReal;

  procedure RandomCC(Wr,Wi: in out Vector; P: in out Poly; RFac,IFac: in Flt := One) is
  begin
    Random(0,Wr,Wi,P,RFac,IFac);
  end RandomCC;

  procedure Random(NSimple: in Natural; Wr,Wi: in out Vector; P: in out Poly; RFac,IFac: in Flt := One) is
  begin
    RandomRoots(NSimple,Wr,Wi,RFac,IFac);
    Roots2Poly(Wr,Wi,P);
  end Random;

  procedure RandomPoly(U,V: in out Vector; P: in out Poly; UFac,VFac: in Flt := Three) is
  begin
    RandomRoots(U,UFac);
    RandomRoots(V,VFac);
    for K in P'Range loop SetZero(P(K)); end loop;
    Assign(One,P(0));
    for I in U'Range loop
      QMultiply(U(I),V(I),Scal_Zero,Scal_Zero,P);
    end loop;
  end RandomPoly;

  procedure RandomSymPoly(U: in out Vector; P: in out Poly; Fac: in Flt := Two) is
    --- Fac=2 corresponds to roots on the unit circle
  begin
    RandomRoots(U,Fac);
    for K in P'Range loop SetZero(P(K)); end loop;
    Assign(One,P(0));
    for K in U'Range loop
      QMultiply(U(K),Scal_One,Scal_Zero,Scal_Zero,P);
    end loop;
  end RandomSymPoly;

  ------------------------------------- Newton stuff
  ------------------------------------- Newton stuff

  procedure Center(V: in out Vec2) is
  begin
    Center(V(1));
    Center(V(2));
  end Center;

  procedure Add(U: in Vec2; V: in out Vec2) is
  begin
    Add(U(1),V(1));
    Add(U(2),V(2));
  end Add;

  procedure Sub(U: in Vec2; V: in out Vec2) is
  begin
    Sub(U(1),V(1));
    Sub(U(2),V(2));
  end Sub;

  procedure Neg(A: in out Mat2) is
  begin
    Neg(A(1,1));
    Neg(A(1,2));
    Neg(A(2,1));
    Neg(A(2,2));
  end Neg;

  procedure Center(A: in out Mat2) is
  begin
    Center(A(1,1));
    Center(A(1,2));
    Center(A(2,1));
    Center(A(2,2));
  end Center;

  procedure Copy(A: in Mat2; B: in out Mat2) is
  begin
    Copy(A(1,1),B(1,1));
    Copy(A(1,2),B(1,2));
    Copy(A(2,1),B(2,1));
    Copy(A(2,2),B(2,2));
  end Copy;

  procedure Add(A: in Mat2; B: in out Mat2) is
  begin
    Add(A(1,1),B(1,1));
    Add(A(1,2),B(1,2));
    Add(A(2,1),B(2,1));
    Add(A(2,2),B(2,2));
  end Add;

  procedure Add(I: in Integer; A: in out Mat2) is
    R: constant Flt := Flt(I);
  begin
    Add(Scal(R),A(1,1));
    Add(Scal(R),A(2,2));
  end Add;

  procedure Prod(A,B: in Mat2; C: in out Mat2) is
  begin
    Prod(A(1,1),B(1,1),C(1,1));
    AddProd(A(1,2),B(2,1),C(1,1));
    Prod(A(1,1),B(1,2),C(1,2));
    AddProd(A(1,2),B(2,2),C(1,2));
    Prod(A(2,1),B(1,1),C(2,1));
    AddProd(A(2,2),B(2,1),C(2,1));
    Prod(A(2,1),B(1,2),C(2,2));
    AddProd(A(2,2),B(2,2),C(2,2));
  end Prod;

  procedure InvDet(A: in Mat2; S,Tmp: in out Scalar) is
  begin
    Prod(A(1,1),A(2,2),Tmp);
    Prod(A(1,2),A(2,1),S);
    Sub(S,Tmp);
    Inv(Tmp,S);
  end InvDet;

  procedure Quot(B: in Vec2; A: in Mat2; X: in out Vec2) is
    Tmp: Scalar renames X(1);
    S: Scalar;
  begin
    InvDet(A,S,Tmp);
    Prod(A(1,2),B(2),X(1));
    Neg(X(1));
    AddProd(A(2,2),B(1),X(1));
    Mult(S,X(1));
    Prod(A(2,1),B(1),X(2));
    Neg(X(2));
    AddProd(A(1,1),B(2),X(2));
    Mult(S,X(2));
  end Quot;

  procedure Quot(B,A: in Mat2; X: in out Mat2) is
    Tmp: Scalar renames X(1,1);
    S: Scalar;
  begin
    InvDet(A,S,Tmp);
    for J in 1 .. 2 loop
      Prod(A(1,2),B(2,J),X(1,J));
      Neg(X(1,J));
      AddProd(A(2,2),B(1,J),X(1,J));
      Mult(S,X(1,J));
      Prod(A(2,1),B(1,J),X(2,J));
      Neg(X(2,J));
      AddProd(A(1,1),B(2,J),X(2,J));
      Mult(S,X(2,J));
    end loop;
  end Quot;

  function MaxNorm(V: Vec2) return Radius is
  begin
    return MaxNorm(V(1))+MaxNorm(V(2));
  end MaxNorm;

  function MaxNorm(A: Mat2) return Radius is
  begin
    return RMax(MaxNorm(A(1,1))+MaxNorm(A(2,1)),MaxNorm(A(1,2))+MaxNorm(A(2,2)));
  end MaxNorm;

  procedure BallAt0(R: in Flt; V: in out Vec2) is
  begin
    BallAt0(R,V(1));
    BallAt0(R,V(2));
  end BallAt0;

  procedure BallAt0(R: in Flt; A: in out Mat2) is
  begin
    BallAt0(R,A(1,1));
    BallAt0(R,A(1,2));
    BallAt0(R,A(2,1));
    BallAt0(R,A(2,2));
  end BallAt0;

  procedure Add(I: in Integer; S: in out Scalar) is
  begin
    Add(Scal(Flt(I)),S);
  end Add;

  procedure Find01(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Newton1;
    procedure Find is new Find0 (TF => TF);
  begin
    Find(S,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end Find01;

  procedure FindZero1(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Newton1;
    procedure Find is new FindZero (F => F, DF => DF);
  begin
    Find(S,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end FindZero1;

  procedure Find02(V: in out Vec2; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Newton2;
    procedure Find is new Find0 (TF => TF);
  begin
    Find(V,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end Find02;

  procedure FindZero2(V: in out Vec2; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Newton2;
    procedure Find is new FindZero (F => F, DF => DF);
  begin
    Find(V,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end FindZero2;

end PolyRoots;
