with Strings, Globals, Flts.Plotting;
use Strings, Globals;

pragma Elaborate_All (Strings,Globals,Flts.Plotting);

package body ScalVectors.Phi is

  function FindMu(Wr: Vector) return Radius is
    --- assume that some Wr(J)>1
    W: Flt;
    E1,E2: Flt := One;
  begin
    for I in Wr'Range loop
      W := Approx(Wr(I));
      if W>E1 then
        E2 := E1;
        E1 := W;
      elsif W>E2 then
        E2 := W;
      end if;
    end loop;
    if E1=E2 then raise Sorry; end if;
    return Half*(E1+E2);
  end FindMu;

  function Check_Unstable(Phi: Matrix; Mu,NPD: Radius) return Boolean is
    --- check if NPD*||(Phi-Mu)^{-1}||<1
    Pow4: constant Positive := 4;
    Eps: constant Radius := 1.0E-999;
    R: constant Radius := -(NegOne/NPD+Eps); -- less than 1/NPD
    S: constant Scalar := Scal(Mu);
    A: Matrix(Phi'Range(1),Phi'Range(2));
  begin
    Copy(Phi,A);
    for I in A'Range(1) loop Sub(S,A(I,I)); end loop;
    Invert(A);
    return CheckNorm2(A,R,Pow4);
  end Check_Unstable;

  procedure IndexG(Vr,Vi: in Vector; S: in out Scalar) is
    --- S := G(V,V)
    L: constant Positive := Vr'Last/2;
    Tmp: Scalar;
  begin
    SetZero(S);
    for I in 1 .. L loop
      Prod(Vr(I),Vi(I+L),Tmp);
      Add(Tmp,S);
      Prod(Vr(I+L),Vi(I),Tmp);
      Sub(Tmp,S);
    end loop;
    Mult(-Two,S,Tmp);
  end IndexG;

  procedure IndexG(J: in Positive; Vr,Vi: in Matrix; S: in out Scalar) is
    -- S := G(V(*,J),V(*,J))
    L: constant Positive := Vr'Last(1)/2;
    Tmp: Scalar;
  begin
    SetZero(S);
    for I in 1 .. L loop
      Prod(Vr(I,J),Vi(I+L,J),Tmp);
      Add(Tmp,S);
      Prod(Vr(I+L,J),Vi(I,J),Tmp);
      Sub(Tmp,S);
    end loop;
    Mult(-Two,S,Tmp);
  end IndexG;

  procedure PosIndexG(Wr: in Vector; Wi: in out Vector; Vr: in Matrix; Vi: in out Matrix) is
    --- change signs of Wi(J) and Vi(*,J), if necessary, to get nonnegative index
    S: Scalar;
  begin
    for J in Vr'Range(2) loop
      IndexG(J,Vr,Vi,S);
      if Sign(S)<0 then
        Neg(Wi(J));
        for I in Vr'Range(1) loop Neg(Vi(I,J)); end loop;
      end if;
      if Verbosity>1 then
        Show2(Strng(J,3) & " Wr,Wi",Approx(Wr(J)),Approx(Wi(J)),False);
        Show1("  Index",Approx(S));
      end if;
    end loop;
  end PosIndexG;

  -----------------------------

  function Low_Eta(Wr,Wi: Vector; Vr,Vi: Matrix) return F_Intervals is
    L: constant Positive := Wr'Last;

    procedure Make0(Eta: in Scalar; F: out F_Interval) is
    begin
      F.Num_Lower := Inf(Eta);
      F.Num_Upper := Sup(Eta);
      F.Den := 1;
      F.Finite := True;
    end Make0;

    Pi2,S,Eta1,Eta2: Scalar;
    F: F_Intervals(1 .. 2*L);
  begin
    ArcCos(Zero_Scalar,S); -- Pi/2
    Prod(Four,S,Pi2);      -- 2*Pi
    for J in 1 .. L loop
      if MaxNorm(Wr(J))<One then
        ArcCos(Wr(J),S);
        Quot(S,Pi2,Eta1);
        Diff(One_Scalar,Eta1,Eta2);
        if Sign(Wi(J))<0 then Swap(Eta1,Eta2); end if;
        declare
          F1: F_Interval renames F(2*J-1);
          F2: F_Interval renames F(2*J);
        begin
          Make0(Eta1,F1);
          Make0(Eta2,F2);
          IndexG(J,Vr,Vi,S);
          F1.Index := Sign(S);
          F2.Index := -F1.Index;
        end;
      end if;
    end loop;
    Check_Intervals(F);
    if Verbosity>4 then Show1("low_eta",F); end if;
    return F;
  end Low_Eta;

  function Separating_Eta(Low,High: F_Intervals) return FltVec is
    Eps: constant Radius := 1.0E-6;
    Z: Parity;
    L: Integer;
    R: Flt;
    V: F_Intervals(1 .. Low'Length+High'Length) := Low & High;
  begin
    Gaps(V,L,Z);
    declare
      G: FltVec(1 .. L+Z);
    begin
      if Z=1 then G(1) := Zero; end if;
      for I in 1 .. L loop
        R := Half*(V(I).Num_Lower+V(I).Num_Upper);
        if abs(R-Half)<Eps then
          G(I+Z) := Half;
        else
          G(I+Z) := R;
        end if;
      end loop;
      return G;
    end;
  end Separating_Eta;

  ----------- plotting

  procedure Plot_Circle(Name: in String; Steps: in Positive := 256) is
    use Flts.Plotting, Flt_EF;
    TwoPi: constant Flt := 6.283185307179586476925286766559;
    DT: constant Flt := TwoPi/Flt(Steps);
    T: Flt;
    F: File_Type;
  begin
    Create(F,Out_File,Name);
    for N in 0 .. Steps loop
      T := Flt(N)*DT;
      Plot2(F,Cos(T),Sin(T));
    end loop;
    Close(F);
  end Plot_Circle;

  procedure Plot_Eigen_Circle(Stem: in String; Wr: in Vector) is
    use Flts.Plotting, Flt_EF;
    Name: constant String := Stem & ".wr.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X)>One then
        Plot2(F,X,Zero);
        Plot2(F,One/X,Zero);
      else
        Y := Sqrt(One-X*X);
        Plot2(F,X, Y);
        Plot2(F,X,-Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Circle;

  procedure Plot_Eigen_Pos(Stem: in String; Wr,Wi: in Vector) is
    use Flts.Plotting;
    Name: constant String := Stem & ".wpos.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X) <= One then
        Y := Approx(Wi(J));
        Plot2(F,X,Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Pos;

  procedure Plot_Eigen_Neg(Stem: in String; Wr,Wi: in Vector) is
    use Flts.Plotting;
    Name: constant String := Stem & ".wneg.xy";
    X,Y: Flt;
    F: File_Type;
  begin
    Show0("Writing " & Name);
    Create(F,Out_File,Name);
    for J in Wr'Range loop
      X := Approx(Wr(J));
      if abs(X) <= One then
        Y := Approx(Wi(J));
        Plot2(F,X,-Y);
      end if;
    end loop;
    Close(F);
  end Plot_Eigen_Neg;

  function Count_Eigen_Zero(Wr: Vector) return Natural is
    Count: Natural := 0;
  begin
    for J in Wr'Range loop
      if abs(Approx(Wr(J)))>One then Count := Count+1; end if;
    end loop;
    return Count;
  end Count_Eigen_Zero;

  procedure Plot_Eigen_Zero(Stem: in String; Wr: in Vector) is
  begin
    if Count_Eigen_Zero(Wr)>0 then
      declare
        use Flts.Plotting;
        Name: constant String := Stem & ".wzero.xy";
        X: Flt;
        F: File_Type;
      begin
        Create(F,Out_File,Name);
        for J in Wr'Range loop
          X := Approx(Wr(J));
          if abs(X)>One then
            Plot2(F,X,Zero);
            Plot2(F,One/X,Zero);
          end if;
        end loop;
        Close(F);
      end;
    end if;
  end Plot_Eigen_Zero;

end ScalVectors.Phi;
